#!/bin/bash

##### Dash Masternode Status Console v0.91-beta - Derekt Development, 2025.
### DMSC is a Dash (Digital Cash) masternode/network status console with audible alerts.
##
# DMSC should work on pretty much any GNU based *nix system without installing anything.
# Requirements: awk, bash (v4.2+), bc, coreutils, Dash Core, grep, iproute2, jq
# Commands used: awk, bash, bc, dash-cli, grep, jq, sleep, ss, tr
#
# Usage: Modify dashConf variable. Dash Core must be configured for RPC. Main/test-net is detected automatically. ^C to exit.
#
# Logic:
#	 Every loopInterval we check for a new block, waiting up to 2 seconds for each RPC and ss to return.
#	 Alert if no RPC data in over loopInterval * noRpcCountMax.
#	 Alert if no new block discovered in over loopInterval * blockStallCountMax.
#        Alert if any unknown masternode related error.
#	 Alert if PoSe score indicates a penalty or ban.
#	 PoSe score is green if the situation is the same or getting worse.
#	 PoSe score is blue with audible alert muted when conditions are improving.
#        ICN is red with audible alert when masternode is active without inbound connecting nodes.
#
# Example mainnet output:
#
# blk: 2221740  diff: 109125713.  MN: Not configured.  DKG: 0 12  MNC: 2046 299  ICN: 9  pool: 3 7216/300000000  mem: 39232/262144  load: 0.08  0.07  0.08  2025-02-13 06:17:44
# 
# Explanation:
#
# Block height, difficulty, MN state, Distributed Key Generation <active next>, MN count <regular evo>, Inbound Connecting
# Nodes, mempool transaction count memory usage/max, memory used/allocated, system load averages, wall clock time. If
# InstantSend or ChainLocks are disabled, that will be displayed as well. Difficulty displayed is limited to 10 characters to
# conserve horizontal screen space.
##
# The idea behind this tool is to have a simple and secure means to monitor masternode/network health. Blocking I/O is limited
# in order to provide a more real-time state per line of output. The goal is to provide maximum useful information per line of
# output. Any error states are logged to dmsc.log. DMSC has minimal requirements, and does not depend on email, which is
# unreliable and has significant privacy concerns.
#
# Default setting for loopInterval is 1 for maximum 'reasonable' output granularity. deciseconds are supported to assist in
# Dash Core debugging. blockStallCountMax is 3600 (1h), which is many times the block target (2.6m). All PoW schemes routinely
# see high variance. This is a good thing, but makes it impossible to provide a proper early warning system for block discovery
# without relying on external data.
#
# Setting loopInterval below 1, and/or decreasing dashCli timeout will increase system overhead and result in diminishing
# returns, depending on the exact hardware or virtualized environment. Exercise caution. Time estimates in output are prone to
# extreme dilation effects based on load levels combined with alert delays. Wall clock time is on each line of output to help
# gauge this effect.
#
# My bash skills are pretty rusty (and ancient), so I'm sure there are bugs. We don't work with data from external sources,
# so any bugs shouldn't be a security issue. As always, assess your IT situations carefully! This code is best effort, and
# offers no guarantees. The most ideal way to use it is combined with other scripts and/or monitoring systems specific to
# your setup and needs. Effort has been made to reduce fork/exec()'s and memory usage, but I'm sure further optimizations
# without additional dependencies are possible.
#
# TODO: Improve output? More advanced logging/alerting? Display network/support devnets? Display additional metrics? Add
#       additional checks? Evo specific features?
#
#       Improve unexpected first phase error parsing by checking Dash Core code.
#
#       A significant design concern is, PoSe score will drop each block after a penalty, regardless of the state of service
#       provided. This means a ban can essentially come out of nowhere without additional checks. Sadly, even if we had them,
#       it seems unlikely we could easily prevent a ban. LLMQ participation doesn't allow for daemon restarts, and it seems
#       Dash Core isn't very robust when it comes to preventing strange issues. I've seen all sorts of crazy things happen
#       when testing on underpowered systems, which is great for testing DMSC, but definitely not good for the security and
#       integrity of the network. It should be possible to make dashd more robust in this regard, but we need to properly
#       identify some of these edge cases. They seem to largely stem from disk I/O bottlenecks or hangs. It would probably
#       be helpful to check debug.log for clues in certain situations, but it will require quite a bit more testing to see
#       what exactly should be done.
#
#       That said, after many hours of consideration, testing, and studying RPC output, at this point we're really close to
#       a proper v1. Keep in mind, this started as a personal script to help assess the viability of potential MN hosting
#       environments. It has since evolved into Dash's most advanced anti-fragility tool to date, but we still have a long
#       way to go. Now that we have a solid base, the next step is to fully assess the PoSe design. It really doesn't make
#       sense for v2 to be scraping debug.log, etc, and potentially trying to patch things up. From what I've seen so far,
#       PoSe definitely needs to be reworked. In the process, we should be thinking a lot about the additional overhead,
#       potential supply risks, etc, from Evo, and how that could impact base layer functionality/fundamentals.
##
# Derekt Development believes in strong base layer fundamentals. Innovation is important, but it should never come at the
# expense of being proper, hard digital money. Dash should be as anti-fragile and future-proof as possible!
#
# Shoutout to xk for creating Masternode Zeus, and for providing code and bash optimization tips! :)
#
# Feedback/patches/requests welcome! If you appreciate these sorts of efforts, but can't afford to donate, please do still
# send a trivial amount of DASH to show support. DASH: XyJAMs7LSTYg9FxJe7kVg7YMxD28TFAM3y 

dashConf="/etc/dash.conf"
if [ ! -f "$dashConf" ]; then
    echo "Dash Core configuration not found!"
    exit
fi
# loopInterval 0.1 (1 decisecond) is the lowest supported value.
loopInterval=1
blockStallCountMax=3600
noRpcCountMax=1
# Display first phase data every loopInterval. Local masternode/block related data is displayed when there's a new block.
# This is useful for debugging, and for calibrating maximum reasonable output granularity in a new environment. Start by
# setting loopInterval to 0.1. If you don't average ~10 lines/second under low load, increase by a decisecond until you see
# the expected output.
verbose=1
# If loopInterval is below 1, this can create a large dmsc.log rather quickly.
logVerbose=0

dmscVersion="v0.91-beta"
dashCli="timeout 2 dash-cli -conf=$dashConf"
previousBlock=0
blockStallCount=0
blockDiscoveryStalled=0
noRpcCount=0
previousPoseScore=0
previousDifficulty=0

# ANSI codes
darkBlue='\e[34m'
teal='\e[36m'
normal='\e[0m'
purple='\e[35m'
gold='\e[33m'
darkGreen='\e[32m'
darkRed='\e[31m'

mnPort=9999
[[ -r "$dashConf" ]] && grep -q "^testnet=1" "$dashConf" && mnPort=19999

# Support deciseconds to assist in debugging Dash Core.
loopSubSec=${loopInterval%.*}
if ((loopSubSec == 0)); then
  loopDs=${loopInterval#??}
else
  loopDs=$((loopInterval * 10))
fi

alert(){
  local i
  for ((i=1; i<=$1; i++))
    do echo -en '\a';sleep 0.25; done
}

countLines(){
  local lines
  if (($#==1)); then
      while IFS= read junk; do ((lines++)); done < $1
  else
      while IFS= read junk; do ((lines++)); done
  fi
  echo $lines
}

echo -e "${darkBlue}D${teal}ash ${darkBlue}M${teal}asternode ${darkBlue}S${teal}tatus ${darkBlue}C${teal}onsole${normal} $dmscVersion..."

while :; do
 while :; do
   block="$($dashCli getblockcount 2>&1)"
   # Check block height every loopInterval for lack of RPC data and block discovery stalling. Run local masternode and block related checks when block count increments.
   if [[ -n "$block" ]]; then
    noRpcCount=0
    memPoolInfo="$($dashCli getmempoolinfo 2>&1)"
    if [[ -n "$memPoolInfo" ]]; then
      memPoolSize=$(jq .size <<< $memPoolInfo)
      memPoolUsage=$(jq .usage <<< $memPoolInfo)
      maxMemPool=$(jq .maxmempool <<< $memPoolInfo)
    fi
    memoryInfo="$($dashCli getmemoryinfo 2>&1)"
    if [[ -n "$memoryInfo" ]]; then
      memoryUsed=$(jq .locked.used <<< $memoryInfo)
      memoryTotal=$(jq .locked.total <<< $memoryInfo)
    fi
    masternodeCount="$($dashCli masternode count 2>&1)"
    if [[ -n "$masternodeCount" ]]; then
      mncEnabled="$(jq .detailed.regular.enabled <<< $masternodeCount) $(jq .detailed.evo.enabled <<< $masternodeCount)"
    fi
    inboundNodes="$(timeout 2 ss -o state established '( sport = :'$mnPort' )' 2>&1)"
    if [[ -n "$inboundNodes" ]]; then
      inboundNodesCount="$(tail -n+2 <<<$inboundNodes|countLines)"
      if [[ -z "$inboundNodesCount" ]]; then
        inboundNodesCount=0
      fi
    fi
    load="$(awk '{print $1,$2,$3}' /proc/loadavg)"
    printf -v date '%(%Y-%m-%d %H:%M:%S)T\n' -1

    if ((verbose == 1)) && ((previousBlock != 0)) && ((block != 0)) && ((blockDiscoveryStalled != 1)); then
      echo -en "${purple}blk:${normal} $block${normal}  ${purple}MNC:${normal} $mncEnabled  ${purple}ICN:${normal} $inboundNodesCount  ${purple}pool:${normal} $memPoolSize $memPoolUsage${darkBlue}/${normal}$maxMemPool  ${purple}mem:${normal} $memoryUsed${darkBlue}/${normal}$memoryTotal  ${purple}load:${normal} $load  $date"
      if ((logVerbose == 1)); then
        echo -en "${purple}blk:${normal} $block${normal}  ${purple}MNC:${normal} $mncEnabled  ${purple}ICN:${normal} $inboundNodesCount  ${purple}pool:${normal} $memPoolSize $memPoolUsage${darkBlue}/${normal}$maxMemPool  ${purple}mem:${normal} $memoryUsed${darkBlue}/${normal}$memoryTotal  ${purple}load:${normal} $load  $date">>dmsc.log
      fi
    fi

    # Clean up any unexpected first phase errors and display them. Should be improved per TODO.
    getBlockErr="$(grep -i error <<<$block|tr '\n' ' ')"
    if [[ -n "$getBlockErr" ]]; then
	echo -en "${gold}$getBlockErr${normal}  ${purple}ICN:${normal} $inboundNodesCount  ${purple}load:${normal} $load  $date"
        echo -en "${gold}$getBlockErr${normal}  ${purple}ICN:${normal} $inboundNodesCount  ${purple}load:${normal} $load  $date">>dmsc.log
	alert 3
    elif ((block == 0)); then
        echo -en "Chain sync initializing... ${purple}MNC:${normal} $mncEnabled  ${purple}ICN:${normal} $inboundNodesCount  ${purple}pool:${normal} $memPoolSize $memPoolUsage${darkBlue}/${normal}$maxMemPool  ${purple}mem:${normal} $memoryUsed${darkBlue}/${normal}$memoryTotal  ${purple}load:${normal} $load  $date"
    # getblockcount returned without error, and previousBlock is 0, so we just started.
    elif ((previousBlock == 0)); then
        previousBlock=$block
        break
    elif ((block == previousBlock)); then
        blockStallCount=$(($blockStallCount+1))
        if ((blockStallCount > blockStallCountMax)); then
          echo -en "${teal}BLOCK DISCOVERY STALLED! > ~$(($blockStallCount * $loopDs))ds${normal}  ${purple}blk:${normal} $block${normal}  ${purple}MNC:${normal} $mncEnabled ${purple}ICN:${normal} $inboundNodesCount ${purple}pool:${normal} $memPoolSize $memPoolUsage${darkBlue}/${normal}$maxMemPool  ${purple}mem:${normal} $memoryUsed${darkBlue}/${normal}$memoryTotal  ${purple}load:${normal} $load  $date"
          echo -en "${teal}BLOCK DISCOVERY STALLED! > ~$(($blockStallCount * $loopDs))ds${normal}  ${purple}blk:${normal} $block${normal}  ${purple}MNC:${normal} $mncEnabled  ${purple}ICN:${normal} $inboundNodesCount  ${purple}pool:${normal} $memPoolSize $memPoolUsage${darkBlue}/${normal}$maxMemPool  ${purple}mem:${normal} $memoryUsed${darkBlue}/${normal}$memoryTotal  ${purple}load:${normal} $load  $date">>dmsc.log
          alert 2
          blockDiscoveryStalled=1
        fi
    elif ((block > previousBlock)) && ((blockDiscoveryStalled == 1)); then
	echo -en "Block discovery resuming. Discovery took > ~$(($blockStallCount * $loopDs))ds!   $date"
	echo -en "Block discovery resuming. Discovery took > ~$(($blockStallCount * $loopDs))ds!   $date">>dmsc.log
	previousBlock=$block
	blockStallCount=0
	blockDiscoveryStalled=0
	break
    elif ((block > previousBlock)); then
        previousBlock=$block
        blockStallCount=0
        break
    fi
   else
     noRpcCount=$(($noRpcCount+1))
     if ((noRpcCount > noRpcCountMax)); then
       echo -en "${teal}SYSTEM LIKELY OVERLOADING! No RPC data in > ~$(($noRpcCount * $loopDs))ds!${normal}  ${purple}ICN:${normal} $inboundNodesCount  ${purple}load:${normal} $load  $date"
       echo -en "${teal}SYSTEM LIKELY OVERLOADING! No RPC data in > ~$(($noRpcCount * $loopDs))ds!${normal}  ${purple}ICN:${normal} $inboundNodesCount  ${purple}load:${normal} $load  $date">>dmsc.log
       alert 3
     fi
   fi
   sleep $loopInterval
 done

  echo -en "${purple}blk:${normal} $block  "

  difficulty="$($dashCli getdifficulty 2>&1)"
  if [[ -n "$difficulty" ]]; then
    didDifficultyReduce=$(bc <<< "$difficulty < $previousDifficulty")
    if ((didDifficultyReduce == 1)); then
      echo -en "${purple}diff:${normal} ${darkBlue}${difficulty:0:10}${normal}   "
    else
      echo -en "${purple}diff:${normal} ${darkGreen}${difficulty:0:10}${normal}   "
    fi
    previousDifficulty=$difficulty
  fi

  mnStatus="$($dashCli masternode status 2>&1)"
  if [[ -n "$mnStatus" ]]; then
   # If numLines is 25, then we have an active MN, so we can avoid several pattern matches of our largest output this cycle.
   numLines="$(countLines <<<$mnStatus)"
   if ((numLines == 25)); then
    mnActive=1
    poseScore="$(jq .dmnState.PoSePenalty <<< $mnStatus)"
    if ((poseScore == 0)); then
     echo -en "${purple}MN:${normal} Status nominal.  "
    elif ((poseScore > 0)); then
     echo -en "${purple}MN:${normal} ${darkRed}PoSe PENALTY${normal} "
     echo -en "${purple}blk:${normal} $block\t${purple}MN:${normal} ${darkRed}PoSe PENALTY${normal} ">>dmsc.log
     if ((poseScore < previousPoseScore)); then
        echo -en "${darkBlue}$poseScore${normal}  "
	echo -en "${darkBlue}$poseScore${normal}  ${purple}MNC:${normal} $mncEnabled\t${purple}ICN:${normal} $inboundNodesCount\t${purple}pool:${normal} $memPoolSize $memPoolUsage${darkBlue}/${normal}$maxMemPool\t${purple}mem:${normal} $memoryUsed${darkBlue}/${normal}$memoryTotal\t${purple}load:${normal} $load  $date">>dmsc.log
     else
        echo -en "${darkGreen}$poseScore${normal}  "
	echo -en "${darkGreen}$poseScore${normal}  ${purple}MNC:${normal} $mncEnabled\t${purple}ICN:${normal} $inboundNodesCount\t${purple}pool:${normal} $memPoolSize $memPoolUsage${darkBlue}/${normal}$maxMemPool\t${purple}mem:${normal} $memoryUsed${darkBlue}/${normal}$memoryTotal\t${purple}load:${normal} $load  $date">>dmsc.log
        alert 4
     fi
    fi
    previousPoseScore=$poseScore
   else
    mnActive=0 
    mnNotSetup="$(grep -c 'does not run' <<<$mnStatus)"
    mnWaiting="$(grep -c WAITING_FOR_PROTX <<<$mnStatus)"
    mnError="$(grep Error <<<$mnStatus|awk -F\" '{print $4}')"
    mnBanned="$(grep -c POSE_BANNED <<<$mnStatus)"
    if ((mnNotSetup == 1)); then
     echo -en "${purple}MN:${normal} Not configured.  "
    elif ((mnWaiting == 1)); then
     echo -en "${purple}MN:${normal} Awaiting ProTx.  "
    elif [[ -n "$mnError" ]]; then
     echo -en "${gold}$mnError${normal}  "
     echo -en "${purple}blk:${normal} $block\t${gold}$mnError${normal}  ${purple}MNC:${normal} $mncEnabled\t${purple}ICN:${normal} $inboundNodesCount\t${purple}pool:${normal} $memPoolSize $memPoolUsage${darkBlue}/${normal}$maxMemPool\t${purple}mem:${normal} $memoryUsed${darkBlue}/${normal}$memoryTotal\t${purple}load:${normal} $load  $date">>dmsc.log 
     alert 4
    elif ((mnBanned == 1)); then
     echo -en "${purple}MN:${normal} ${darkRed}PoSe BANNED!${normal}  "
     echo -en "${purple}blk:${normal} $block\t${purple}MN:${normal} ${darkRed}PoSe BANNED!${normal}  ${purple}MNC:${normal} $mncEnabled\t${purple}ICN:${normal} $inboundNodesCount\t${purple}pool:${normal} $memPoolSize $memPoolUsage${darkBlue}/${normal}$maxMemPool\t${purple}mem:${normal} $memoryUsed${darkBlue}/${normal}$memoryTotal\t${purple}load:${normal} $load  $date">>dmsc.log
     alert 5
    fi
   fi
  fi

  quorumDkgInfo="$($dashCli quorum dkginfo 2>&1)"
  if [[ -n "$quorumDkgInfo" ]]; then
    quorumActiveDkgs=$(jq .active_dkgs <<< $quorumDkgInfo)
    quorumNextDkg=$(jq .next_dkg <<< $quorumDkgInfo)
    echo -en "${purple}DKG:${normal} $quorumActiveDkgs $quorumNextDkg  "
  fi

  if [[ -n "$mncEnabled" ]]; then
    echo -en "${purple}MNC:${normal} $mncEnabled "
  fi

  if [[ -n "$inboundNodes" ]]; then
    if ((inboundNodesCount == 0)) && ((mnActive == 1)); then
      echo -en "${purple}ICN:${normal} ${darkRed}$inboundNodesCount${normal}  "
      alert 4
    else
      echo -en "${purple}ICN:${normal} $inboundNodesCount  "
    fi
  fi

  if [[ -n "$memPoolInfo" ]]; then
    echo -en "${purple}pool:${normal} $memPoolSize $memPoolUsage${darkBlue}/${normal}$maxMemPool  "
  fi

  if [[ -n "$memoryInfo" ]]; then
    echo -en "${purple}mem:${normal} $memoryUsed${darkBlue}/${normal}$memoryTotal  "
  fi

  sporkShow="$($dashCli spork show 2>&1)"
  if [[ -n "$sporkShow" ]]; then
    instantsendEnabled=$(jq .SPORK_2_INSTANTSEND_ENABLED <<< $sporkShow)
    chainlocksEnabled=$(jq .SPORK_19_CHAINLOCKS_ENABLED <<< $sporkShow)
    if (("$instantsendEnabled" != 0)); then
      echo -en "${purple}IS:${normal} ${darkRed}DISABLED!${normal}  "
    fi
    if (("$chainlocksEnabled" != 0)); then
      echo -en "${purple}CL:${normal} ${darkRed}DISABLED${normal}  "
    fi
  fi

  echo -en "${purple}load:${normal} $load  $date"
  sleep $loopInterval
done

### EOF
